var pjs = module.exports = {};

var errors = require('./errors.js');
var utils = require('./utils.js');

var initialized = false;
var workers = [];

function init(options) {
	if (initialized) {
		throw new errors.InvalidOperationError(errors.messages.CONSECUTIVE_INITS);
	}

	options = options || {};
	var cpus = navigator.hardwareConcurrency || 1;
	var maxWorkers = options.maxWorkers || cpus;
	var workersCount = Math.min(maxWorkers, cpus);

	var wCode = function (event) {};
	var blob = new Blob(["onmessage = " + wCode.toString()]);
	var blobURL = window.URL.createObjectURL(blob);
	while (workersCount--) {
		var worker = new Worker(blobURL);
		workers.push(worker);
	}

	var config = {
  	get workers () {
  		return workers.length;
  	}
  };

  utils.getter(pjs, 'config', config);

	initialized = true;
}

function terminate() {
	if (!initialized) {
		throw new errors.InvalidOperationError(errors.messages.TERMINATE_WITHOUT_INIT);
	}

	workers.forEach(function (w) { w.terminate(); });
	workers = [];
	delete pjs.config;

	initialized = false;
};

pjs.init = init;
pjs.terminate = terminate;